from simunetcore.serializer import Serializer
import numpy as np
    
class Model(Serializer):
    """Model base class."""  

    def __init__(self, name=""):
        """
        Creates a new Model object.

        Parameters
        ----------
        name : str
            The name of the model.
 
        """
       
        Serializer.__init__(self)
        
        # Meta Data
        self.name = name        # Name of the model
        self.description = ""   # Description of the model
        self.dim = "unknown"    # Dimension of the model ["unknown", "0D", "1D", "2D", "3D"]
        self.domain = "unknown" # Domain of the model [e.g. "unknown", "VT", "PLM", "LCA"]
        
        # Enpoint of the the model
        self.site = "local"     # Computation site ["local", "remote"]       
        self.threadsafe = False # Supports threading [True, False]

        # Model specific members
        self.params = {}        # Parameters of the model
        self.vars = {}          # State variables of the model (dependent in time and space)
        self.u = {}             # Input vectors of the model
        self.y = {}             # Output vectors of  the model
        self.t = [0]            # Time base of the values (dim(t) = dim(var))
        self.adj = []           # Internal adjacencies of the model
  
    def __get_value(self, port, port_id, t):
        
        result = []
        for var_id in port[port_id]["vars"]:
            result.append(self.get_var(var_id, t))
        return result
        
    def __set_value(self, port, port_id, value):
        
        if len(value) != len(port[port_id]["vars"]):
            raise ValueError("Shape mismatch")
        for var_id in port[port_id]["vars"]:
            index = port[port_id]["vars"].index(var_id)
            if isinstance(value[index], np.ndarray):
                self.vars[var_id]["value"] = value[index].tolist() 
            else:
                self.vars[var_id]["value"] = value[index]
            
    def __get_port_vars(self, port):
        
        var_set = set()
        for port_id in port:
            var_set = var_set.union(port[port_id]["vars"])
        return var_set
    
    def get_params(self):
        """
        Returns the parameters of the model.

        Returns
        -------
        value : set of str
            A set of ids of the parameters of the model.
    
        """
        return set(self.vars.keys())

    def get_param(self, param_id):
        """
        Returns the value of the parameter with id `param_id`.

        Returns
        -------
        value : Any
            The value of the parameter with id `param_id`.

        Parameters
        ----------
        param_id : str
            The id of the parameter to retrieve.
    
        """
       
        return self.params[param_id]["value"]

    def set_param(self, param_id, value):
        """
        Sets the value of the parameter with id `param_id`.

        Parameters
        ----------
        param_id : str
            The id of the parameter to set.
        value : Any
            The value to be stored under the id `param_id`.
    
        """
       
        if isinstance(value, np.ndarray):
            self.params[param_id]["value"] = value.tolist() 
        else: 
            self.params[param_id]["value"] = value

    def get_vars(self):
        """
        Returns the variables of the model.

        Returns
        -------
        value : set of str
            A set of ids of the variables of the model.
    
        """
        
        return set(self.vars.keys())

    def get_var(self, var_id, t=None):
        """
        Returns the value of the variable with id `var_id`.
        If `t` is given the value is interpolated for the given time steps.

        Returns
        -------
        value : Any
            The value of the variable with id `var_id`.

        Parameters
        ----------
        var_id : str
            The id of the variable to retrieve.
        t : float or list of float, optional
            The time steps for which the result is to be interpolated. 
            Deafult = None.
    
        """
        
        if t is None:
            return self.vars[var_id]["value"]  
        else:
            return np.interp(t, self.get_t(), self.vars[var_id]["value"]).tolist()

    def set_var(self, var_id, value):
        """
        Sets the value of the variable with id `var_id`.

        Parameters
        ----------
        var_id : str
            The id of the variable to set.
        value : Any
            The value to be stored under the id `var_id`.
    
        """
       
        if isinstance(value, np.ndarray):
            self.vars[var_id]["value"] = value.tolist()  
        else:
            self.vars[var_id]["value"] = value

    def get_t(self):
        """
        Returns the current time frame of the state variables of the model.

        Returns
        -------
        value : float or list of float
            The time frame of the state variables of the model.
        """       
        return self.t
    
    def set_t(self, value):
        """
        Sets the time frame for the state variables of the model.

        Parameters
        ----------
        value : float, list of float or np.ndarray
            The time frame to be stored.
    
        """
        
        if isinstance(value, np.ndarray):
            self.t = value.tolist()  
        else:
            self.t = value
        
    def get_u(self, u_id, t=None):
        """
        Returns the values of the variables of the input port with id `u_id`.
        If `t` is given the values are interpolated for the given time steps.

        Returns
        -------
        value : list
            The values of the variables of the port with id `u_id`.

        Parameters
        ----------
        u_id : str
            The id of the input port.
        t : float or list of float, optional
            The time steps for which the result is to be interpolated.  
            Deafult = None.
    
        """
        
        return self.__get_value(self.u, u_id, t)
    
    def set_u(self, u_id, value):
        """
        Sets the values of the variables of the input port with id `u_id`.

        Parameters
        ----------
        u_id : str
            The id of the input port.
        value : list
            The value to be stored for the variables of the port 
            with the id `port_id`.
    
        """
       
        if isinstance(value, np.ndarray):
            self.__set_value(self.u, u_id, value.tolist())
        else:
            self.__set_value(self.u, u_id, value)
        
    def get_u_vars(self):
        """
        Returns the input variables of the model.

        Returns
        -------
        value : set of str
            A list of ids of the the input variables of the model.
    
        """
       
        return self.__get_port_vars(self.u)

    def get_y(self, y_id, t=None):
        """
        Returns the values of the variables of the output port with id `y_id`.
        If `t` is given the values are interpolated for the given time steps.

        Returns
        -------
        value : list
            The values of the variables of the port with id `y_id`.

        Parameters
        ----------
        y_id : str
            The id of the output port.
        t : float or list of float, optional
            The time steps for which the result is to be interpolated.  
            Deafult = None.
    
        """
        
        return self.__get_value(self.y, y_id, t)
    
    def set_y(self, y_id, value):
        """
        Sets the values of the variables of the output port with id `y_id`.

        Parameters
        ----------
        y_id : str
            The id of the output port.
        value : list
            The value to be stored for the variables of the port 
            with the id `port_id`.
    
        """
        
        if isinstance(value, np.ndarray):
            self.__set_value(self.y, y_id, value.tolist())
        else:
            self.__set_value(self.y, y_id, value)
    
    def get_y_vars(self):
        """
        Returns the output variables of the model.

        Returns
        -------
        value : set of str
            A list of ids of the the output variables of the model.
    
        """
        
        return self.__get_port_vars(self.y)
    
    def compute(self, *args, **kwargs):
        """
        Computes the state variables.

        Paramters
        ---------
        *args : Any
            A list of arguments used by derived classes.
        **kwargs : dict {str : Any}
            A dictionary of keyword : argument pairs used by derived classes.
    
        """
        if self.get_vars() == self.get_u_vars():
            return

        raise NotImplementedError("Method 'compute' is not implemented in 'Model'")



import gzip
import base64
import io
from datetime import datetime

def compress(content):
    """
    Compresses the content using gzip and base64 encode.
    Raises a TypeError if argument is None.

    Parameters
    ----------
    content : str
        The content to be compressed.

    Returns
    -------
    compressed_text : str
        A string containing the base64 encoded and compressed content.

    Raises
    ------
    `ValueError` if argument `content` is None.
    
    """

    if content is None:
        raise ValueError("Argument is None")

    # Convert content to string
    text = str(content)

    # Compress text using gzip
    stream_out = io.BytesIO()
    with gzip.GzipFile(fileobj = stream_out, mode = "w") as zip_file:
        zip_file.write(text.encode())

    # Encode text using base64
    compressed_text = base64.b64encode(stream_out.getvalue())

    # Return the result
    return compressed_text.decode()
    
def decompress(content):
    """
    Decompresses the base64 encoded content using gzip and base64.
    Raises `TypeError` if argument is not of type str.

    Parameters
    ----------
    content : str
        The base64 encoded and gzipped content to be decompressed.

    Returns
    -------
    decompressed_text : str
        A string containing the decoded and decompressed text.

    Raises
    ------
    `TypeError` if argument `content` is not a str.

    """

    if not isinstance(content, str):
        raise TypeError("String argument expected")
        
    # Convert to string
    text = str(content)  
    # Decode the text from base64
    text = base64.b64decode(text.encode())

    # Compressde text using gzip
    stream_in = io.BytesIO()
    stream_in.write(text)
    stream_in.seek(0)
    with gzip.GzipFile(fileobj = stream_in, mode = "rb") as zip_file:
        decompressed_text = zip_file.read()

    # Return the result
    return decompressed_text.decode()
 
def current_time():
    """
    Returns the current utc time in isoformat.

    Returns
    -------
    utc_now : str
        A string containing the utc time in isoformat.

    """
    utc_now = datetime.utcnow().isoformat()[:-3]
    return utc_now


from simunetcore.model import Model
from simunetcore import exception
import requests

class RESTModel(Model):
    """REST model class."""  
    
    def __init__(self, name=""):
        """
        Creates a new Model object.

        Parameters
        ----------
        name : str
            The name of the model.
 
        """
       
        Model.__init__(self, name)
        
        # Remote compute endpoint of the model
        self.endpoint = {
            "url": "", 
            "timeout": 300,
            "compress": False,
            "header": {}
        } 
    
    def compute(self, invocation_type="compute", **kwargs):
        """
        Computes the state variables.

        Paramters
        ---------
        invocation_type : str, optional
            The type of function invocation. Must be on of "warmup" 
            or "compute".
            Default = "compute".
        **kwargs : dict {str : Any}
            A dictionary of keyword : argument pairs used by derived classes.
    
        Raises
        ------
        `RemoteFunctionError` if soemthing went wrong on the remote site.
        `ParameterError` if parameters are not correct.

       """
        
        # Get function
        function = self.endpoint["url"]
        
        # Get compression settings
        use_compression = self.endpoint["compress"]

        # Create payload
        if invocation_type == "warmup":
            body = {}
            
        elif invocation_type == "compute":
            # Create and compress body
            body = self.dumps(compress=use_compression)
            
        else:
            raise exception.ParameterError(
                "Parameter 'invocation_type' must be 'compute' or 'warmup'")

        # Create headers dict and request
        headers_dict = self.endpoint["header"]
        headers_dict["sn-use-compression"] = str(use_compression)

        # Post the request
        response = requests.post(
            function, 
            data=body, 
            timeout=self.endpoint["timeout"], 
            headers = headers_dict)

        # Check the result
        if invocation_type == "compute" and response.status_code == 200:                  
            # load the data from the result
            self.loads(response.text, decompress=use_compression)
            
        elif invocation_type == "warmup" and response.status_code == 204:
            pass
        
        else:
            # Raise error if something went wrong
            raise exception.RemoteFunctionError(
                function=function, 
                invocation_type=invocation_type, 
                result="statusCode {}, {}".format(
                    response.status_code,
                    response.text))



"""
Simulation Network simunet
"""

import datetime
from simunetcore.serializer import Serializer
from simunetcore.model import Model
from simunetcore.rest import RESTModel
from simunetcore.mqtt import MQTTModel
from simunetcore.aws import AWSModel
from simunetcore.waveform import Waveform
from simunetcore.factory import Factory

__all__ = ["Serializer", "Model", "RESTModel", "MQTTModel", "AWSModel", "Waveform", "Factory"]
__title__ = "simunetcore"
__version__ = "0.0.13"
__license__ = "MIT License"
__copyright__ = "(c){} core simunet(r) team and contributors".format(datetime.date.today().year)

# Set default logging handler to avoid "No handler found" warnings.
import logging
from logging import NullHandler

logging.getLogger(__name__).addHandler(NullHandler())



import inspect
import types
from simunetcore import Model
from simunetcore import MQTTModel
from simunetcore import AWSModel
from simunetcore import RESTModel
from simunetcore import logger
        
class Factory(dict):
    """The simunet factory class."""

    def __init__(self):
        """
        Creates a new factory object and 
        registers the core simunet models 
        (AWSModel, MQTTModel and RESTModel).
 
        """
        super().__init__()
       
        self["AWSModel"] = AWSModel
        self["MQTTModel"] = MQTTModel
        self["RESTModel"] = RESTModel

    def register(self, code, namespace = "simunetcore.factory"):
        """
        Rergisters a piece of code as a model in the factory 
        in the given namespace.

        Parameters
        ----------
        code : str
            The source code of the model class. 
        namespace : str, optional
            The namespace used for the model.
            It is registered as <namespace>.<class_name>
            For example Registering a class `TestModel` 
            in the namespace `my.models`  leads to the 
            dict entry: {'Test': <class 'my.models.TestModel'>}.
            Default = 'simunetcore.factory'.

        """            
        # create a factory module
        factory_module = types.ModuleType(namespace)
        factory_module.__loader__ = self

        exec(code, factory_module.__dict__)
        members_list = inspect.getmembers(factory_module, inspect.isclass)
        for item in members_list:
            if issubclass(item[1], Model) and item[1] is not Model:
                self[item[0]] = item[1]
                logger.info("Registered {}: {}".format(item[0], item[1]))



import matplotlib.ticker as ticker
import numpy as np

def plot_convergence(
        figure, convergence, epsilon,
        col_0="C0", col_1="C1"):
    """
    Creates a convergence plot.

    Paramters
    ---------
    figure : Figure
        The matplotlib Figure, which holds all plot elements.
    convergence : list
        The convergence data for each frame.
    epsilon : float
        The tolerance value.
    col_0 : str, optional
        The color code used for the tolerance data scatter points.
        Default = "C0".
    col_1 : str, optional
        The color code used for the tolerance data line.
        Default = "C1".

    """

    # Create axis
    ax = figure.add_subplot()
    
    # Segments start
    ax.axvline(x=-0.5, color="k", linestyle=":", linewidth=1)

    # Plot convergence progress starting with the first iteration
    iter_start = 0
    for conv_frame in convergence:
        # Retrieve the elements with the maximum error 
        # for each frame in the convergence data 
        conv_data = list(max(conv_iter)[0] for conv_iter in conv_frame)
        num_elements = len(conv_data)
        iter_data = np.linspace(
            iter_start, iter_start + num_elements - 1, num_elements)

        # Plot markers for the retrieved values
        ax.scatter(
            iter_data, conv_data, color=col_1, marker="o", alpha=0.4)

        # Now plot a line and mark only the first and the last element
        markevery = (0, max(len(conv_frame) - 1, 1))
       
        ax.plot(
            iter_data, conv_data, color=col_0, marker="o",
            linestyle="-", markevery=markevery,
            label="$\\varphi(y_i)$" if iter_start == 0 else None)

        # Set the new iteration start value
        iter_start += num_elements

    # Plot the segment limits again starting with the first iteration
    iter_start = 0
    for conv_frame in convergence:
        conv_data = list(max(conv_iter)[0] for conv_iter in conv_frame)
        num_elements = len(conv_data)
        iter_start += num_elements
        
        # Draw a vertical line
        ax.axvline(
            x=iter_start - 0.5, color="k", 
            linestyle=":", linewidth=1, alpha=1)

    # Plot the tolerance value epsilon as a red horizontal line
    ax.plot(
        [0, iter_start], [epsilon, epsilon],
        color="r", linestyle="-", linewidth=1,
        label="$\epsilon={}$".format(epsilon))
    ax.axhline(y=epsilon, color="r", linestyle="-", linewidth=1)

    # Set y-axis to logarithmic scale
    ax.set_yscale("log")

    # Set axis labels
    ax.set_xlabel("Iteration [-]")
    ax.set_ylabel("$\\varphi(y_i)$ [-]")
    
    # Enable grid
    ax.grid(which="major", axis="y")
    if len(convergence) > 0:
        ax.xaxis.set_major_locator(ticker.MaxNLocator(integer=True))
        ax.xaxis.set_minor_locator(ticker.AutoMinorLocator())
    else:
        ax.xaxis.set_major_locator(ticker.FixedLocator([0]))
        ax.set_xticklabels([""])

    # Add the legend
    ax.legend(loc=1)

def plot_iterations(
        figure, statistics, epsilon, 
        col_0="C0", col_1="C1", col_2="C3"):
    """
    Creates an iterations plot.

    Paramters
    ---------
    figure : Figure
        The matplotlib Figure, which holds all plot elements.
    statistics : list
        The iteraton results for each frame.
    epsilon : float
        The tolerance value.
    col_0 : str, optional
        The color code used for finished frames.
        Default = "C0".
    col_1 : str, optional
        The color code used for restarted frames.
        Default = "C1.
    col_2 : str, optional
        The color code used for canceled frames.
        Default = "C3".

    """

    # Create axis
    ax = figure.add_subplot()

    # Create data lists
    num_frame = [row[1] for row in statistics]
    error = [row[7] for row in statistics]

    # Retrieve the finished, reastarted an canceled frames from the data
    num_frame_finished = [row[1] for row in statistics if row[2] == "Finished"]
    num_iter_frame_finished = [row[6] for row in statistics if row[2] == "Finished"]

    num_frame_restarted = [row[1] for row in statistics if row[2] == "Restarted"]
    num_iter_frame_restarted = [row[6] for row in statistics if row[2] == "Restarted"]

    num_frame_canceled = [row[1] for row in statistics if row[2] == "Canceled"]
    num_iter_frame_canceled = [row[6] for row in statistics if row[2] == "Canceled"]

    # Plot the restarted frames
    if len(num_frame_restarted):
        ax.bar(
            num_frame_restarted, num_iter_frame_restarted, 
            color=col_1, alpha=0.4, label="Restarted")
       
    # Plot the finished frames
    ax.bar(
        num_frame_finished, num_iter_frame_finished, 
        color=col_0, alpha=0.4, label="Finished")
    
    # Plot the canceled frames
    if len(num_frame_canceled):
        ax.bar(
            num_frame_canceled, num_iter_frame_canceled, 
            color=col_2, alpha=0.4, label="Canceled")

    # Set axis labels
    ax.set_ylabel("Iterations [-]")
    ax.set_xlabel("Frames [-]")

    # Set axismajor locators
    if len(statistics) > 0:
        ax.xaxis.set_major_locator(ticker.MaxNLocator(integer=True))
    else:
        ax.xaxis.set_major_locator(ticker.FixedLocator([0]))
        ax.set_xticklabels([""])

    # Enable grid
    ax.grid(axis="y")

    # Twin axis
    ax2 = ax.twinx() 

    # Plot the convergence results
    ax2.plot(
        num_frame, error, color=col_0, 
        linestyle="-", marker="o",
        label="$\\varphi(y_i)$")

    # Set axis labels and style
    if np.any(error):
        ax2.set_yscale("log")
    ax2.set_ylabel("$\\varphi(y_i)$ [-]")

    # Plot the tolerance value epsilon as a red horizontal line
    ax2.axhline(y=epsilon, color="r", linestyle="-", linewidth=1.0)
    ax2.plot(
        num_frame, np.full(len(num_frame), epsilon), 
        color="r", linestyle="-", linewidth=1.0, 
        label="$\epsilon={}$".format(epsilon))
    
    # Get handles and labels
    handles1, labels1 = ax.get_legend_handles_labels()
    handles2, labels2 = ax2.get_legend_handles_labels()

    # Add legend
    ax2.legend(handles1+handles2, labels1+labels2, loc=1)


from simunetcore import mqtt
from simunetcore import aws
from threading import Thread
import json
import traceback

class Invoker(object):
    """MQTT Invoker base class."""  
    
    def __init__(self, provider, simulation_id, model_id, function):
        """
        Creates a new Invoker object.
        The parameters are used to subscribe to the mqtt broker on:
        `/simunet/{provider}/{simulation_id}/{model_id}/{function}/task`

        Parameters
        ----------
        provider : str
            The provider id within simunet.
        simulation_id : str
            The simulation id to serve.
        model_id : str
            The model id to serve.
        function : str
            The function name.
 
        """
        
        # Store properties
        self.provider = provider
        self.simulation_id = simulation_id
        self.model_id = model_id
        self.function = function
        
        # MQTT communication client
        self.mqtt_client = None
        
        # Request topic
        self.topic_request  = "/simunet/{}/{}/{}/{}/task".format(
            provider, simulation_id, model_id, function)     
    
    def __on_message(self, client, userdata, msg):
        
        # Create thread
        t = Thread(
            target=self.compute_model, 
            args=(msg.topic, msg.payload.decode()))
        
        # Fire & forget
        t.start()

    def compute_model(self, topic, payload):
        """
        Computes the associated model.

        Paramters
        ---------
        topic : str
            The subscripted mqtt topic.
        payload : str
            The payload of the mqtt message.
    
        """

        raise NotImplementedError(
            "Method 'compute_model' is not implemented in 'Invoker'")
         
    def connect_mqtt(self, mqtt_endpoint):
        """
        Connects to a mqtt broker and subscribes to:
        `/simunet/provider/simulation_id/model_id/function/task`

        Paramters
        ---------
        mqtt_endpoint : dict{str: Any}
            The endpoint dictionary containing the mqtt parameters.
            {
                "broker": "mqtt.broker.url",
                "port": 1883,
                "transport": "tcp",
                "tls": False,
                "username": "user",
                "password": "passwd"
            }

        See Also
        --------
        paho.mqtt library : simunet uses paho for mqtt communication.

        """        
        # Create mqtt client
        self.mqtt_client = mqtt.createMQTTClient(
            mqtt_endpoint, "{}[{}|{}|{}]".format(
                self.provider, self.simulation_id, self.model_id, self.function))
        
        # Attach callbacks
        self.mqtt_client.on_message = self.__on_message     
        # Subscribe to topic
        self.mqtt_client.subscribe(self.topic_request)        
        # Start listener loop
        self.mqtt_client.loop_start()

    def disconnect_mqtt(self):
        """
        Disconnects from the mqtt broker.

        """
        
        # Stop listener loop
        self.mqtt_client.loop_stop()
        # Unsubscribe
        self.mqtt_client.unsubscribe(self.topic_request)
        # Disconnect
        self.mqtt_client.disconnect()
                
class AWSInvoker(Invoker):
    """AWS Invoker class."""  
    
    def __init__(self, simulation_id):
        """
        Creates a new AWSInvoker object.
        The parameter is used to subscribe to the mqtt broker on
        `/simunet/invoker/{simulation_id}/+/+/task` so that requests 
        for all values for `model_id` and `function` are catched.

        Parameters
        ----------
        simulation_id : str
            The simulation id to serve.
 
        """
        
        Invoker.__init__(
            self, provider="invoker", simulation_id=simulation_id, 
            model_id="+", function="+")
            
        # AWS lambda client
        self.lambda_client = None
                         
    def compute_model(self, topic, payload):
        """
        Computes the associated model.

        Paramters
        ---------
        topic : str
            The subscripted mqtt topic.
        payload : str
            The payload of the mqtt message.
    
        """
        
        # Get function name from topic
        topic_list = topic.split("/")
        function = topic_list[-2]
        
        # Call Lambda function
        response = self.lambda_client.invoke(
            FunctionName=function, 
            InvocationType='RequestResponse',
            Payload=payload
        )

        # Get the result
        result = response["Payload"].read()
        
        # Publish the result
        topic_list[-1] = "result"
        topic_response = "/".join(topic_list)
        self.mqtt_client.publish(topic_response, result)   

    def connect_aws(self, aws_endpoint):
        """
        Connects to AWS lambda.

        Parameters
        ----------
        aws_endpoint : dict{str: Any}
            The endpoint dictionary containing the aws parameters.
            {
                "aws_id": "aws_id",
                "aws_key": "aws_key",
                "region": "eu-central-1",
                "max_attempts": 0,
                "read_timeout": 900,
                "connect_timeout": 60 
            }

        See Also
        --------
        boto3 library : simunet uses boto3 for aws communication.

        """
        
        # Create boto client
        self.lambda_client = aws.createLambdaClient(aws_endpoint)
        
class LocalInvoker(Invoker):
    """Local Invoker class."""  
    
    def __init__(self, model, provider, function):
        """
        Creates a new LocalInvoker object.
        The parameters 'provider' and 'function' are used to 
        subscribe to the mqtt broker on:
        `/simunet/{provider}/+/+/{function}/task` so that requests 
        for all values for `simulation_id` and `model_id` are catched.

        Parameters
        ----------
        model : simunet.Model
            The associated simunet model.
        provider : str
            The provider id within simunet.
        function : str
            The function name.

        """
        
        Invoker.__init__(
            self, provider=provider, simulation_id="+", 
            model_id="+", function=function)
        
        # Store the connected model
        self.model = model
        
    def compute_model(self, topic, payload):
        """
        Computes the associated model.

        Paramters
        ---------
        topic : str
            The subscripted mqtt topic.
        payload : str
            The payload of the mqtt message.
    
        """
        
        try:         
            # Get simulation id an model id name from topic
            topic_list = topic.split("/")
            simulation_id = topic_list[-4]
            model_id = topic_list[-3]

            # Task disctionary
            task = json.loads(payload)
            # Result dictionary
            result = {}

            # Call function
            if task:
                self.model.loads(task["body"])
                self.model.compute(
                    simulation_id = simulation_id,
                    model_id = model_id)
                result["statusCode"] = 200
                result["body"] = self.model.dumps(indent = True)
            else:
                result["statusCode"] = 204
                result["body"] = ""
                
        except:
            # Set error code and error messgae
            result["statusCode"] = 500
            result["body"] = traceback.format_exc()
            
        finally:
            # Publish the result
            topic_list[-1] = "result"
            topic_response = "/".join(topic_list)
            self.mqtt_client.publish(topic_response, json.dumps(result, indent=True))
        